
/*
 * Copyright (c) 1997 Carter Bullard
 * All applicable rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation is restricted to personal use only.  Use, sale
 * or retransmission of this software for commercial purposes, 
 * including but not limited to use as a commerical product or
 * in support of a commercial endeavor requires licensing from Carter
 * Bullard.
 *
 * CARTER BULLARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS, IN NO EVENT SHALL CARTER BULLARD BE LIABLE FOR ANY
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

/*
 * Copyright (c) 1993, 1994 Carnegie Mellon University.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted, 
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear
 * in supporting documentation, and that the name of CMU not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  
 * 
 * CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
 * CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 */

/*
 *
 * policy - parse policy configuration and report argus data that
 *          does not conform to this policy.   no policy will report
 *          all data.
 *
 *          this is the principle argus client that uses the client
 *          libraries which provide code to read argus archive files,
 *          attach to remote argus servers, parse policy configuration
 *          files, parse argus data streams and calls data specific
 *          subroutines that are provided in this file.
 *
 *          the principle funciton of this client is to print out
 *          argus data, and its intended use is to review argus data.
 *
 */

#define ARGUS_CLIENT

#if !defined(SOLARIS)
#include <stdlib.h>
#include <unistd.h>
#endif

#include <compat.h>

#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#ifndef SOLARIS
#include <strings.h>
#endif
#include <string.h>
#include <time.h>

#include <sys/socket.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>

#include <arpa/inet.h>
#include <netdb.h>

#include <policy.h>

#include <cons_out.h>
#include <argus_parse.h>

#include <pcap.h>
#include <interface.h>
#include <addrtoname.h>
#include <nametoaddr.h>

#define IPPROTOSTR 99
char *ip_proto_string [IPPROTOSTR] = {"ip", "icmp", "igmp", "ggp",
   "ipnip", "st", "tcp", "ucl", "egp", "igp", "bbn-rcc-mon", "nvp-ii",
   "pup", "argus", "emcon", "xnet", "chaos", "udp", "mux", "dcn-meas",
   "hmp", "prm", "xns-idp", "trunk-1", "trunk-2", "leaf-1", "leaf-2",
   "rdp", "irtp", "iso-tp4", "netblt", "mfe-nsp", "merit-inp", "sep",
   "3pc", "idpr", "xtp", "ddp", "idpr-cmtp", "tp++", "il", "unas",
   "unas", "unas", "unas", "unas", "unas", "unas", "unas", "unas",
   "unas", "unas", "unas", "unas", "unas", "unas", "unas", "unas",
   "unas", "unas", "unas", "any", "cftp", "any", "sat-expak", "kryptolan",
   "rvd", "ippc", "any", "sat-mon", "visa", "ipcv", "cpnx", "cphb", "wsn",
   "pvp", "br-sat-mon", "sun-nd", "wb-mon", "wb-expak", "iso-ip", "vmtp",
   "secure-vmtp", "vines", "ttp", "nsfnet-igp", "dgp", "tcf", "igrp",
   "ospfigp", "sprite-rpc", "larp", "mtp", "ax.25", "ipip", "micp",
   "aes-sp3-d", "etherip", "encap",
};

char *icmptypestr[ICMP_MAXTYPE + 1] = {
   "ECR", "   ", "   ", "UR", "SRC", "RED",
   "   ", "   ", "ECO", "   ", "   ", "TIM",
   "PAR", "TST", "TSR", "IRQ", "IRR", "MAS",
   "MSR",
};    


char buf[1024];
char *get_tcp_string ();
char *get_icmp_string ();
char *get_ip_string ();
char *get_udp_string ();
char *get_man_init_string ();
char *get_man_stat_string ();

char *appOptstring = NULL;
extern int major_version, minor_version;
extern struct tm tm_startime, tm_lasttime;

extern u_char src_ttl, dst_ttl;

init () {}
void argus_parse_complete () {}
void clientTimeout () {}
parse_arg (argc, argv)
int argc;
char**argv;
{}

void
usage (ptr)
char *ptr;
{
   fprintf (stderr, "usage: %s [-bchmnEIMNORTWX] ", ptr);
   fprintf (stderr, "[-C access-file] ");
   fprintf (stderr, "[-d debug-level]\n");
   fprintf (stderr, "              [-r input_file] ");
   fprintf (stderr, "[-w output_file] ");
   fprintf (stderr, "[-F file] [-P port]\n              ");
   fprintf (stderr, "[-t time_range] [-S argus_server] expression.\n");
   fprintf (stderr, "options: b - dump packet-matching code.\n");
   fprintf (stderr, "         c - print packet and byte counts.\n");
   fprintf (stderr, "         C - specify Cisco access-file.\n");
   fprintf (stderr, "         h - print help.\n");
   fprintf (stderr, "         m - print MAC addresses.\n");
   fprintf (stderr, "         n - don't convert numbers to names.\n");
   fprintf (stderr, "         r - read Argus input file.\n");
   fprintf (stderr, "         t - specify time_range for search.\n         ");
   fprintf (stderr, "      format:    ");
   fprintf (stderr, "timeSpecification[-timeSpecification]\n");
   fprintf (stderr, "                          timeSpecification: ");
   fprintf (stderr, "[mm/dd[/yy].]hh[:mm[:ss]]\n");
   fprintf (stderr, "                                             ");
   fprintf (stderr, " mm/dd[/yy]\n");
   fprintf (stderr, "         u - print msecs timestamps (*)\n");
   fprintf (stderr, "         w - write Argus output file.\n");
   fprintf (stderr, "         x - flip ports on input (*)\n");
   fprintf (stderr, "         P - specify remote argus port.\n");
   fprintf (stderr, "         S - specify remote argus host.\n");
   fprintf (stderr, "         E - print ESTABLISHED transactions.\n");
   fprintf (stderr, "         I - print extended ICMP status.\n");
   fprintf (stderr, "         F - use file to define filter expression.\n");
   fprintf (stderr, "         M - print MULTIROUTE transactions.\n");
   fprintf (stderr, "         N - print NORMAL_CLOSED transactions.\n");
   fprintf (stderr, "         O - print transactions with ip OPTIONS.\n");
   fprintf (stderr, "         R - print RESET transactions.\n");
   fprintf (stderr, "         T - print TIMEDOUT transactions.\n");
   fprintf (stderr, "         W - print WINDOW_SHUT transactions.\n");
   fprintf (stderr, "         X - print PKTS_RETRANS transactions.\n");
   exit(1);
}

#define FLAGSTRLEN   5

process_man (ptr)
struct WriteStruct *ptr;
{
   int vc = 0;
   char *str;
   char strbuf[256];

   switch (ptr->status & (INIT|STATUS|CLOSE)) {
      case   INIT:  str = get_man_init_string (ptr); break;
      case STATUS:  str = get_man_stat_string (ptr); break;
      case  CLOSE:  str = get_man_stat_string (ptr); break;
   }

   if (str) {
      switch (ptr->status & (INIT|STATUS|CLOSE)) {
         case   INIT: strcat (str, "INT"); ; break;
         case STATUS: strcat (str, "STA"); ; break;
         case  CLOSE: strcat (str, "CLO"); ; break;
      }
   
      (void) printf ("%s\n", str);
      fflush (stdout);
   }
}


process_frag (ptr)
struct WriteStruct *ptr;
{
   int vc = 0;
   char *str;
   char strbuf[256];
   struct fragWriteStruct *frag = &ptr->ws_ip_frag;

   switch (ptr->status & (UDPPROTO|ICMPPROTO|TCPPROTO)) {
      case UDPPROTO:  str = get_udp_string (ptr); break;
      case ICMPPROTO: str = get_icmp_string (ptr); break;
      case TCPPROTO:  str = get_tcp_string (ptr); break;
      default:        str = get_ip_string (ptr); break;
   }

   bcopy ((char *) "frag", (char *)&str[19], 4);

   if (mflag) vc += 17;
   str[vc + 50] = ' ';
   sprintf (&str[vc + 69], " %-5d ", frag->frag_id);

   if (cflag) {
      sprintf (strbuf, "pk %2d  ex %4d  ob %4d  max %4d ",
          frag->fragnum, frag->totlen, frag->currlen,
          frag->maxfraglen);
      strcat (str, strbuf);
   }

   strcat (str, "TIM");
   (void) printf ("%s\n", str);
   fflush (stdout);
}

process_tcp (ptr)
struct WriteStruct *ptr;
{
   printf ("%s\n", get_tcp_string (ptr));
}

process_icmp (ptr)
struct WriteStruct *ptr;
{
   printf ("%s\n", get_icmp_string (ptr));
   fflush (stdout);
}

process_udp (ptr)
struct WriteStruct *ptr;
{
   printf ("%s\n", get_udp_string (ptr));
   fflush (stdout);
}

process_ip (ptr)
struct WriteStruct *ptr;
{
   if (ptr->status & FRAG_ONLY)
      process_frag (ptr);
   else
      printf ("%s\n", get_ip_string (ptr));
   fflush (stdout);
}


void print_date(ptr,date)
struct WriteStruct *ptr;
char *date;
{
   if (ptr->status & DETAIL) {
     if (uflag)
       sprintf(date, "%9ld.%-6ld", (unsigned long) ptr->ws_ip.lasttime.tv_sec,
	                          (unsigned long) ptr->ws_ip.lasttime.tv_usec);
     else
       strftime ((char *) date, 128, "%a %m/%d %T", &tm_lasttime);
   } else {
     if (uflag)
       sprintf(date, "%9ld.%-6ld", (unsigned long) ptr->ws_ip.startime.tv_sec,
	                          (unsigned long) ptr->ws_ip.startime.tv_usec);
     else
       strftime ((char *) date, 128, "%a %m/%d %T", &tm_startime);
   }
}

/*
char *interfacestr [INTERFACES] = {"ip", "icmp", "igmp", "ggp",
   "ipnip", "st", "tcp", "ucl", "egp", "igp", "bbn-rcc-mon", "nvp-ii",
*/


char *
get_man_init_string (ptr)
struct WriteStruct *ptr;
{
   int i, ind, vc = 0;
   unsigned int state;
   struct tm *tm;
   unsigned int *netAddr, *netMask;
   char *dstString, *srcString, *protoStr;
   char date [128], fmtstr[256], *blankstring = "     ";
   char *processStr = NULL;

   bzero (fmtstr, 256);
   bzero (buf, sizeof (buf));

   if (uflag)
      sprintf(date, "%9ld.%-6ld", (unsigned long) ptr->ws_init.now.tv_sec,
	                          (unsigned long) ptr->ws_init.now.tv_usec);
   else
      if (tm = localtime ((time_t *)&ptr->ws_init.now.tv_sec))
         strftime ((char *) date, 128, "%a %m/%d %T", tm);

   protoStr =  "man";

   if (mflag) {
      vc = 17;
      strcpy (fmtstr, 
         "%s %17.17s %17.17s       %3s %15.15s %-5.5s     %15.15s %-5.5s ");
   } else
      strcpy (fmtstr, "%s      %3s %15.15s %-5.5s     %15.15s %-5.5s ");

   if (cflag)
      strcat (fmtstr, "%-6d %-6d  %-9d %-9d");

   netAddr  = (unsigned int *) &ptr->ws_init.localnet;
   if (*netAddr)
      while ((*netAddr & 0xff000000) == 0)
         *netAddr <<= 8;

   netMask = (unsigned int *) &ptr->ws_init.netmask;
   if (*netMask)
      while ((*netMask & 0xff000000) == 0)
         *netMask <<= 8;

   srcString = ipaddr_string (netAddr);
   dstString = ipaddr_string (netMask);

   if (mflag)
      sprintf (buf, fmtstr, date,
         blankstring, blankstring, protoStr,
         srcString, blankstring,
         dstString, blankstring,
         0, 0, 0, 0);
   else
      sprintf (buf, fmtstr, date, protoStr,
         srcString, blankstring, dstString, blankstring,
         0, 0, 0, 0);

   return (buf);
}


char *
get_man_stat_string (ptr)
struct WriteStruct *ptr;
{
   int i, ind, vc = 0;
   unsigned int state;
   struct tm *tm;
   long *netAddr, *netMask;
   char *dstString, *srcString, *protoStr;
   char date [128], fmtstr[256], *blankstring = "     ";
   char *processStr = NULL;

   bzero (fmtstr, 256);
   bzero (buf, sizeof (buf));

   if (uflag)
      sprintf(date, "%9ld.%-6ld", (unsigned long) ptr->ws_init.now.tv_sec,
	                          (unsigned long) ptr->ws_init.now.tv_usec);
   else
      if (tm = localtime ((time_t *)&ptr->ws_stat.now.tv_sec))
         strftime ((char *) date, 128, "%a %m/%d %T", tm);

   protoStr =  "man";

   if (mflag) {
      vc = 17;
      strcpy (fmtstr, 
         "%s %17.17s %17.17s       %3s  pkts %5d  drops %5d  ");
   } else
      strcpy (fmtstr, "%s      %3s  pkts %5d  drops %5d  ");

   if (cflag) {
      strcat (fmtstr, "ip %3hd %3hd  t %3hd %3hd  u %3hd %3hd  i %3hd %3hd  f %3hd %3hd  ");

      if (mflag)
         sprintf (buf, fmtstr, date,
            blankstring, blankstring, protoStr,
            srcString, blankstring,
            dstString, blankstring,
            0, 0, 0, 0);
      else
         sprintf (buf, fmtstr, date, protoStr,
            ptr->ws_stat.pktsRcvd, ptr->ws_stat.pktsDrop, 
            ptr->ws_stat.actIPcons, ptr->ws_stat.cloIPcons,
            ptr->ws_stat.actTCPcons, ptr->ws_stat.cloTCPcons,
            ptr->ws_stat.actUDPcons, ptr->ws_stat.cloUDPcons,
            ptr->ws_stat.actICMPcons, ptr->ws_stat.cloICMPcons,
            ptr->ws_stat.actFRAGcons, ptr->ws_stat.cloFRAGcons);
   } else {
      strcat (fmtstr, " flows %8d       ");

      if (mflag)
         sprintf (buf, fmtstr, date,
            blankstring, blankstring, protoStr,
            ptr->ws_stat.pktsRcvd, ptr->ws_stat.pktsDrop, 
            ptr->ws_stat.actIPcons + ptr->ws_stat.cloIPcons +
            ptr->ws_stat.actTCPcons + ptr->ws_stat.cloTCPcons +
            ptr->ws_stat.actUDPcons + ptr->ws_stat.cloUDPcons +
            ptr->ws_stat.actICMPcons + ptr->ws_stat.cloICMPcons +
            ptr->ws_stat.actFRAGcons + ptr->ws_stat.cloFRAGcons);
      else
         sprintf (buf, fmtstr, date, protoStr,
            ptr->ws_stat.pktsRcvd, ptr->ws_stat.pktsDrop, 
            ptr->ws_stat.actIPcons + ptr->ws_stat.cloIPcons +
            ptr->ws_stat.actTCPcons + ptr->ws_stat.cloTCPcons +
            ptr->ws_stat.actUDPcons + ptr->ws_stat.cloUDPcons +
            ptr->ws_stat.actICMPcons + ptr->ws_stat.cloICMPcons +
            ptr->ws_stat.actFRAGcons + ptr->ws_stat.cloFRAGcons);
   }

   return (buf);
}

char *
get_tcp_string (ptr)
struct WriteStruct *ptr;
{
   int i, src_bytes, dst_bytes, src_count, dst_count, ind, vc = 0;
   unsigned int state;
   struct in_addr *srcAddr, *dstAddr;
   struct ether_addr *esrcAddr, *edstAddr;
   char date[128], fmtstr[256];
   char *dstString, *srcString, *protoStr;
   char *edstString, *esrcString;
   char *processStr = NULL;
   unsigned short srcPort, dstPort;
   u_long port, srcSeq, dstSeq, srcDst, dstDst;

   bzero (fmtstr, 256);
   bzero (buf, sizeof (buf));
   bzero (date, 128);

   port = ptr->ws_ip.port;

   print_date (ptr, date);

   state = ptr->status & 
         (SAW_SYN | NORMAL_CLOSE | SAW_SYN_SENT | CON_ESTABLISHED);

   protoStr = (nflag > 1) ? "6  " : "tcp";
   if (mflag) {
      vc = 17;
      strcpy (fmtstr, 
         "%s %17.17s %17.17s       %3s %15.15s.%-5.5s  -  %15.15s.%-5.5s ");
   } else
      strcpy (fmtstr, "%s      %3s %15.15s.%-5.5s  -  %15.15s.%-5.5s ");

   if (cflag)
      strcat (fmtstr, "%-6d %-6d  %-9d %-9d");

   if (state) {
      if (state & NORMAL_CLOSE) {
         fmtstr[29 + vc] = '>';
         processStr = process_state_strings[3];
      } else {
         if (ptr->status & DETAIL) {
            if (state & CLOSE_WAITING) {
               fmtstr[27 + vc] = '<'; fmtstr[29 + vc] = '>';
               processStr = process_state_strings[6];
            } else
            if (state & CON_ESTABLISHED) {
               fmtstr[27 + vc] = '<'; fmtstr[29 + vc] = '>';
               processStr = process_state_strings[2];
            } else
            if (state & SAW_SYN_SENT) {
               fmtstr[27 + vc] = '<';
               processStr = process_state_strings[1];
            } else
            if (state & SAW_SYN) {
               fmtstr[29 + vc] = '>';
               processStr = process_state_strings[0];
            }
         } else {
            fmtstr[29 + vc] = '>';
            if (!(state & (SAW_SYN | SAW_SYN_SENT)))
               fmtstr[27 + vc] = '<';
            processStr = process_state_strings[3];
         }
      }
   }
   
   if (ptr->status & RESET) {
      fmtstr[28 + vc] = '|';
      processStr = process_state_strings[5];
   } 

   if (ptr->status & TIMED_OUT) {
      fmtstr[28 + vc] = 'o';
      processStr = process_state_strings[4];
   }

   strcat (fmtstr, processStr);

   if (ptr->status & PKTS_RETRANS) fmtstr[vc + 3] = '*';
   if (ptr->status & IPOPTIONMASK) {
      switch (ptr->status & IPOPTIONMASK) {
         case SSRCROUTE:    fmtstr[vc + 2] = 'S'; break;
         case LSRCROUTE:    fmtstr[vc + 2] = 'L'; break;
         case TIMESTAMP:    fmtstr[vc + 2] = 'T'; break;
         case SECURITY:     fmtstr[vc + 2] = '+'; break;
         case RECORDROUTE:  fmtstr[vc + 2] = 'R'; break;
         case SATNETID:     fmtstr[vc + 2] = 'N'; break;
         case IPOPTIONMASK: fmtstr[vc + 2] = 'E'; break;
         default:           fmtstr[vc + 2] = 'M'; break;
      }
   }
   if (ptr->status & MULTIADDR) fmtstr[vc + 4] = 'M';
   if (ptr->status & FRAGMENTS) fmtstr[vc + 4] = 'F';
   if (ptr->status & IPROUTED)  fmtstr[vc + 5] = 'R';

   edstAddr = &ptr->ws_ip_phys.etherdst;
   esrcAddr = &ptr->ws_ip_phys.ethersrc;
   srcAddr = &ptr->ws_ip.src; dstAddr = &ptr->ws_ip.dst;
   srcPort = ((unsigned short *) &port)[0];
   dstPort = ((unsigned short *) &port)[1];
   src_count = ptr->ws_ip_tcp.src_count;
   dst_count = ptr->ws_ip_tcp.dst_count;

   if ((ptr->status & SAW_SYN) &&
            !(ptr->status & (SAW_SYN_SENT|CON_ESTABLISHED|RESET))) {
      srcSeq = ptr->ws_ip_inittcp.seq;
      srcDst = ptr->ws_ip_inittcp.addr;
      src_bytes = 0; dst_bytes = 0;
   } else {
      if ((ptr->status & (SAW_SYN | SAW_SYN_SENT)) &&
                         !(ptr->status & (CON_ESTABLISHED))) {
         dstSeq = ptr->ws_ip_inittcp.seq;
         dstDst = ptr->ws_ip_inittcp.addr;
         src_bytes = 0; dst_bytes = 0;
      } else {
         src_bytes = ptr->ws_ip_tcp.src_bytes;
         dst_bytes = ptr->ws_ip_tcp.dst_bytes;
      }
   }

   if (src_bytes < 0) src_bytes = 0;
   if (dst_bytes < 0) dst_bytes = 0;
  
   srcString = ipaddr_string (srcAddr);
   dstString = ipaddr_string (dstAddr);

   esrcString = etheraddr_string ((u_char *) esrcAddr);
   edstString = etheraddr_string ((u_char *) edstAddr);

   if ((srcPort == 20) || ((srcPort >= 6000) && (srcPort < 6010))) {
      int swap27 = 0, swap29 = 0;
      if (fmtstr[27 + vc] == '<') swap27 = 1;
      if (fmtstr[29 + vc] == '>') swap29 = 1;
      if (!(swap27 && swap29)) {
         if (swap27) fmtstr[29 + vc] = '>', fmtstr[27 + vc] = ' ';
         if (swap29) fmtstr[27 + vc] = '<', fmtstr[29 + vc] = ' ';
      }
 
      if (mflag)
         sprintf (buf, fmtstr, date,
            edstString, esrcString, protoStr,
            dstString, tcpport_string(dstPort),
            srcString, tcpport_string(srcPort),
            dst_count, src_count, dst_bytes, src_bytes);
      else
         sprintf (buf, fmtstr, date, protoStr,
            dstString, tcpport_string(dstPort),
            srcString, tcpport_string(srcPort),
            dst_count, src_count, dst_bytes, src_bytes);
   } else
      if (mflag)
         sprintf (buf, fmtstr, date,
            esrcString, edstString, protoStr,
            srcString, tcpport_string(srcPort),
            dstString, tcpport_string(dstPort),
            src_count, dst_count, src_bytes, dst_bytes);
      else
         sprintf (buf, fmtstr, date, protoStr,
            srcString, tcpport_string(srcPort),
            dstString, tcpport_string(dstPort),
            src_count, dst_count, src_bytes, dst_bytes);
 
   return (buf);
}


char *
get_icmp_string (ptr)
struct WriteStruct *ptr;
{
   int i, vc = 0, rev = 0, hlen, srccnt = 0, dstcnt = 0;
   struct ip *oip;
   struct udphdr *ouh;
   struct icmpWriteStruct *icmp;
   char fmtstr[64], *blankstring = "     ", icmptype[32];
   char extendedstring[32];
   char *dstString, *srcString, *protoStr;
   char *edstString, *esrcString;
   char date[128];
   extern char *timestatmp_fmt;
   extern long timestamp_scale, thiszone;
   extern u_long getnetnumber (), ipaddrtonetmask ();
      
   if (ptr) {
      unsigned char  ra_icmp_type, ra_icmp_code;
      unsigned short ra_icmp_data;
      unsigned long  ra_src_addr = 0, ra_dst_addr = 0, ra_gw_addr = 0;

      icmp = &ptr->ws_ip_icmp;
      bzero (icmptype, sizeof (icmptype));
      bzero (buf, sizeof (buf));
      bzero (extendedstring, sizeof (extendedstring));
      bzero (icmptype, sizeof (icmptype));
      bzero (date, 128);

      print_date (ptr, date);

      if (!(ptr->status & (CON_ESTABLISHED | TIMED_OUT))) {
         ra_icmp_type = icmp->type;
         ra_icmp_code = icmp->code;
         ra_icmp_data = icmp->data;
         ra_src_addr =  ntohl(icmp->srcaddr.s_addr);
         ra_dst_addr =  ntohl(icmp->dstaddr.s_addr);
         ra_gw_addr  =  ntohl(icmp->gwaddr.s_addr);

      } else {
         ra_icmp_type = ((unsigned char *) &ptr->ws_ip_udp.src_bytes) [0];
         ra_icmp_code = ((unsigned char *) &ptr->ws_ip_udp.src_bytes) [1];
         ra_icmp_data = ((unsigned short *)&ptr->ws_ip_udp.src_bytes) [1];
         ra_src_addr  =  ntohl(icmp->gwaddr.s_addr);
         ra_dst_addr  =  ntohl(icmp->gwaddr.s_addr);
         ra_gw_addr  =   ntohl(icmp->gwaddr.s_addr);
      }

      if (ra_icmp_type < (unsigned char) (ICMP_MAXTYPE + 1))
         strcpy (icmptype, icmptypestr[ra_icmp_type]);
      else
         strcpy (icmptype, "UNK");

      switch (ra_icmp_type) {
         case ICMP_UNREACH:
            switch (ra_icmp_code) {
               case ICMP_UNREACH_NET:
                  strcat (icmptype, "N");
                  if (icmp->dstaddr.s_addr) {
                     u_long addr = ra_dst_addr;
                     addr &= ipaddrtonetmask (addr);
                     sprintf (extendedstring, "%s", 
                              ipaddr_string (&addr));
                  }
                  break;
               case ICMP_UNREACH_HOST:
                  strcat (icmptype, "H");
                  if (ra_dst_addr)
                     sprintf (extendedstring, "%s", 
                           ipaddr_string (&ra_dst_addr));
                  break;

               case ICMP_UNREACH_PROTOCOL:
                  strcat (icmptype, "O");
                  if (ra_icmp_data && (ra_icmp_data < IPPROTOSTR))
                     sprintf (extendedstring,"%s",
                        ip_proto_string[ra_icmp_data]);
                  break;

               case ICMP_UNREACH_PORT: {
                  char index = icmp->gwaddr.s_addr;
                  strcat (icmptype, "P"); 

                  if ((ra_icmp_data && ((index < IPPROTOSTR)) && (index > 0))
                          && (minor_version > 5)) {
                     sprintf (extendedstring, "%s port %d",
                        ip_proto_string[index],
                           ra_icmp_data);

                  } else if (ra_icmp_data)
                     sprintf (extendedstring, "%d", ra_icmp_data);
                  break;
               }
               case ICMP_UNREACH_NEEDFRAG:
                  strcat (icmptype, "F"); break;
               case ICMP_UNREACH_SRCFAIL:
                  strcat (icmptype, "S"); break;
            }
            break;

         case ICMP_MASKREPLY:
               if (ra_src_addr)
                  sprintf (extendedstring, "0x%08x", ra_src_addr);
            break;

         case ICMP_REDIRECT:
            switch (ra_icmp_code) {
            case ICMP_REDIRECT_NET:
               (void) sprintf (extendedstring, "net %s",
                         ipaddr_string (&ra_gw_addr));
               break;

            case ICMP_REDIRECT_HOST:
               (void) sprintf (extendedstring, "host %s",
                         ipaddr_string (&ra_gw_addr));
               break;

            case ICMP_REDIRECT_TOSNET:
               (void) sprintf (extendedstring, "tosN %s",
                         ipaddr_string (&ra_gw_addr));
               break;

            case ICMP_REDIRECT_TOSHOST:
               (void) sprintf (extendedstring, "tosH %s",
                         ipaddr_string (&ra_gw_addr));
               break;
            }
            break;

         case ICMP_ECHOREPLY:
            rev = 1;
            break;

         case ICMP_TIMXCEED:
               (void) sprintf (extendedstring, "%s",
                         icmp->code ? "reassembly" : "in-transit");
               break;

         case ICMP_PARAMPROB:
         case ICMP_SOURCEQUENCH:
         case ICMP_ECHO:
         case ICMP_TSTAMP:
         case ICMP_TSTAMPREPLY:
         case ICMP_IREQ:
         case ICMP_IREQREPLY:
         case ICMP_MASKREQ:
         default:
            break;
      }
   
      protoStr = (nflag > 1) ? "1   " : "icmp";
      if (rev) {
         edstString = etheraddr_string ((u_char *) &ptr->ws_ip_phys.ethersrc);
         esrcString = etheraddr_string ((u_char *) &ptr->ws_ip_phys.etherdst);
         dstString = ipaddr_string (&ptr->ws_ip.src);
         srcString = ipaddr_string (&ptr->ws_ip.dst);
      } else {
         esrcString = etheraddr_string ((u_char *) &ptr->ws_ip_phys.ethersrc);
         edstString = etheraddr_string ((u_char *) &ptr->ws_ip_phys.etherdst);
         srcString = ipaddr_string (&ptr->ws_ip.src);
         dstString = ipaddr_string (&ptr->ws_ip.dst);
      }
   
      if (mflag) {
         vc = 15;
         sprintf (fmtstr, "%s",
              "%s %17.17s %17.17s      icmp %15.15s %-5.5s  -  %15.15s %-5.5s");
      } else {
         sprintf (fmtstr, "%s",
              "%s     %-4.4s %15.15s %-5.5s  -  %15.15s %-5.5s");
      }
  
      if (rev) {fmtstr[29 + vc] = '<'; fmtstr[31 + vc] = ' ';}
      else     {fmtstr[29 + vc] = ' '; fmtstr[31 + vc] = '>';}
   
      if (cflag) strcat (fmtstr, " %-6d %-6d  %-17.17s ");
   
      if (ptr->status & CON_ESTABLISHED) {
         srccnt = ptr->ws_ip_udp.src_count;
         dstcnt = ptr->ws_ip_udp.dst_count;
         fmtstr[29 + vc] = ' '; fmtstr[31 + vc] = ' ';
         if (srccnt)
            fmtstr[31 + vc] = '>';
         if (dstcnt)
            fmtstr[29 + vc] = '<';
      } else {
         if (rev) dstcnt = 1;
         else     srccnt = 1;
      }

      strcat (fmtstr, " %3.3s");
   
      if (ptr->status & IPOPTIONMASK) {
         switch (ptr->status & IPOPTIONMASK) {
            case SSRCROUTE:    fmtstr[vc + 2] = 'S'; break;
            case LSRCROUTE:    fmtstr[vc + 2] = 'L'; break;
            case TIMESTAMP:    fmtstr[vc + 2] = 'T'; break;
            case SECURITY:     fmtstr[vc + 2] = '+'; break;
            case RECORDROUTE:  fmtstr[vc + 2] = 'R'; break;
            case SATNETID:     fmtstr[vc + 2] = 'N'; break;
            case IPOPTIONMASK: fmtstr[vc + 2] = 'E'; break;
            default:           fmtstr[vc + 2] = 'M'; break;
         }
      }
      if (ptr->status & MULTIADDR) fmtstr[vc + 4] = 'M';
      if (ptr->status & FRAGMENTS) fmtstr[vc + 4] = 'F';
      if (ptr->status & IPROUTED)  fmtstr[vc + 5] = 'R';
  
      if (mflag) {
         if (cflag) {
            sprintf (buf, fmtstr, date, esrcString, edstString,
               srcString, blankstring, dstString, blankstring, 
            srccnt, dstcnt, extendedstring, icmptype);
         } else
            sprintf (buf, fmtstr, date, esrcString, edstString,
               srcString, blankstring, dstString, icmptype,
               icmptype);
      } else {
         if (cflag) {
            sprintf (buf, fmtstr, date, protoStr,
               srcString, blankstring, dstString, blankstring,
               srccnt, dstcnt, extendedstring, icmptype);
         } else
            sprintf (buf, fmtstr, date, protoStr,
               srcString, blankstring, dstString, blankstring,
               icmptype);
      }
   }
   return (buf);
}


char *
get_udp_string (ptr)
struct WriteStruct *ptr;
{
   int i, vc = 0;
   char *blankstring = "*    ";
   char *dstString, *srcString, *udpString;
   char *edstString, *esrcString, *protoStr;
   char date[128], fmtstr[256];
   u_short srcport, dstport;
   extern char *timestatmp_fmt;
   extern long timestamp_scale, thiszone;

   bzero (buf, sizeof (buf));
   bzero (date, 128);
   esrcString = etheraddr_string ((u_char *) &ptr->ws_ip_phys.ethersrc);
   edstString = etheraddr_string ((u_char *) &ptr->ws_ip_phys.etherdst);
   srcString = ipaddr_string (&ptr->ws_ip.src.s_addr);
   dstString = ipaddr_string (&ptr->ws_ip.dst.s_addr);
   srcport = ((u_short *)&ptr->ws_ip.port)[0];
   dstport = ((u_short *)&ptr->ws_ip.port)[1];

   print_date (ptr, date);

   protoStr = (nflag > 1) ? "17 " : "udp";

   if (mflag) {
      vc = 17;
      strcpy (fmtstr, 
         "%s %17.17s %17.17s       %3s %15.15s.%-5.5s  -  %15.15s.%-5.5s ");
   } else
      strcpy (fmtstr, "%s      %3s %15.15s.%-5.5s  -  %15.15s.%-5.5s ");

   if (cflag)
      strcat (fmtstr, "%-6d %-6d  %-9d %-9d");

   if (ptr->status & IPOPTIONMASK) {
      switch (ptr->status & IPOPTIONMASK) {
         case SSRCROUTE:    fmtstr[vc + 2] = 'S'; break;
         case LSRCROUTE:    fmtstr[vc + 2] = 'L'; break;
         case TIMESTAMP:    fmtstr[vc + 2] = 'T'; break;
         case SECURITY:     fmtstr[vc + 2] = '+'; break;
         case RECORDROUTE:  fmtstr[vc + 2] = 'R'; break;
         case SATNETID:     fmtstr[vc + 2] = 'N'; break;
         case IPOPTIONMASK: fmtstr[vc + 2] = 'E'; break;
         default:           fmtstr[vc + 2] = 'M'; break;
      }
   }
   if (ptr->status & MULTIADDR) fmtstr[vc + 4] = 'M';
   if (ptr->status & FRAGMENTS) fmtstr[vc + 4] = 'F';
   if (ptr->status & IPROUTED)  fmtstr[vc + 5] = 'R';

   fmtstr[27 + vc] = (ptr->ws_ip_udp.dst_count) ? '<' : ' '; 
   fmtstr[29 + vc] = (ptr->ws_ip_udp.src_count) ? '>' : ' '; 
   if ((ptr->status & TIMED_OUT)) strcat (fmtstr, "TIM");
   else if (ptr->status & CON_ESTABLISHED) {
      if ((ptr->ws_ip_udp.src_count == 1) && (ptr->ws_ip_udp.dst_count == 1))
         strcat (fmtstr, "ACC");
      else strcat (fmtstr, "CON");
   } else if (ptr->status & UDP_INIT) strcat (fmtstr, "INT");

   if (mflag)
      sprintf (buf, fmtstr, date,
         esrcString, edstString,
         protoStr,
         srcString, udpport_string (srcport), 
         dstString, udpport_string (dstport), 
         ptr->ws_ip_udp.src_count, ptr->ws_ip_udp.dst_count,
         ptr->ws_ip_udp.src_bytes, ptr->ws_ip_udp.dst_bytes);
   else
      sprintf (buf, fmtstr, date, protoStr,
         srcString, udpport_string (srcport),
         dstString, udpport_string (dstport), 
         ptr->ws_ip_udp.src_count, ptr->ws_ip_udp.dst_count,
         ptr->ws_ip_udp.src_bytes, ptr->ws_ip_udp.dst_bytes);

   return (buf);
}


char *
get_ip_string (ptr)
struct WriteStruct *ptr;
{
   int i, vc = 0;
   char *blankstring = "*    ";
   char *dstString, *srcString, *protoStr ;
   char *edstString, *esrcString;
   char date[128], fmtstr[256], protoStrbuf[16];
   extern char *timestatmp_fmt;
   extern long timestamp_scale, thiszone;
   int src_count, dst_count, src_bytes, dst_bytes;
   u_char proto;
   
   bzero (buf, sizeof (buf));
   bzero (date, 128);
   srcString = ipaddr_string (&ptr->ws_ip.src.s_addr);
   dstString = ipaddr_string (&ptr->ws_ip.dst.s_addr);
   esrcString = etheraddr_string ((u_char *) &ptr->ws_ip_phys.ethersrc);
   edstString = etheraddr_string ((u_char *) &ptr->ws_ip_phys.etherdst);
   src_count = ptr->ws_ip_udp.src_count; dst_count = ptr->ws_ip_udp.dst_count;
   src_bytes = ptr->ws_ip_udp.src_bytes; dst_bytes = ptr->ws_ip_udp.dst_bytes;

   print_date (ptr, date);

   if (mflag) {
      vc = 17;
      strcpy (fmtstr, 
            "%s %17.17s %17.17s     %-5.5s%20.20s  -  %20.20s  ");
   } else
      strcpy (fmtstr, "%s     %4s %15.15s        -  %15.15s       ");

   fmtstr[25 + vc] = (dst_count) ? '<' : ' '; 
   fmtstr[27 + vc] = (src_count) ? '>' : ' '; 

   if (cflag)
      strcat (fmtstr, "%-6d %-6d  %-9d %-9d");

   if ((ptr->status & TIMED_OUT)) strcat (fmtstr, "TIM");
   else if (ptr->status & CON_ESTABLISHED) {
      if ((ptr->ws_ip_udp.src_count == 1) && (ptr->ws_ip_udp.dst_count == 1))
         strcat (fmtstr, "ACC");
      else strcat (fmtstr, "CON");
   } else if (ptr->status & UDP_INIT) strcat (fmtstr, "INT");

   if (ptr->status & IPOPTIONMASK) {
      switch (ptr->status & IPOPTIONMASK) {
         case SSRCROUTE:    fmtstr[vc + 2] = 'S'; break;
         case LSRCROUTE:    fmtstr[vc + 2] = 'L'; break;
         case TIMESTAMP:    fmtstr[vc + 2] = 'T'; break;
         case SECURITY:     fmtstr[vc + 2] = '+'; break;
         case RECORDROUTE:  fmtstr[vc + 2] = 'R'; break;
         case SATNETID:     fmtstr[vc + 2] = 'N'; break;
         case IPOPTIONMASK: fmtstr[vc + 2] = 'E'; break;
         default:           fmtstr[vc + 2] = 'M'; break;
      }
   }
   if (ptr->status & MULTIADDR) fmtstr[vc + 4] = 'M';
   if (ptr->status & FRAGMENTS) fmtstr[vc + 4] = 'F';
   if (ptr->status & IPROUTED)  fmtstr[vc + 5] = 'R';

   proto = ptr->ws_ip.port;
   sprintf (protoStrbuf, "%u", proto);

   protoStr = (nflag > 1) ? protoStrbuf :
             proto >= IPPROTOSTR ? "unas" : ip_proto_string[proto];

   if (mflag && cflag)
      sprintf (buf, fmtstr, date, 
            esrcString, edstString, protoStr,
            srcString, dstString, 
            src_count, dst_count,
            src_bytes, dst_bytes);
   else
   if (cflag)
      sprintf (buf, fmtstr, date, protoStr,
            srcString, dstString, 
            src_count, dst_count,
            src_bytes, dst_bytes);
   else
   if (mflag)
      sprintf (buf, fmtstr, date, 
            esrcString, edstString,  protoStr,
            srcString, dstString);
   else
      sprintf (buf, fmtstr, date, protoStr,
            srcString, dstString);

   return (buf);
}

